package net.ae7q.memory;

import java.io.*;
import java.util.*;
// import javax.comm.*;
import gnu.io.*;

//==================================================================================================
/** Ft817calibration implements the data for an FT-817.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
class Ft817base extends Ft___Base {

    //==============================================================================================
    /** ChannelGroup describes the radio's channel groups.
     */
    public enum ChannelGroup implements FtGroup {
        VfoA      ( 0x007D, Ft817options.Subbands.values() ),
        VfoB      ( 0x0203, Ft817options.Subbands.values() ),
        Home      ( 0x0389, Ft817options.HomeBands.values() ),
        Temporary ( 0x03F1, Temporaries.values() ),
        Memory    ( 0x0484,   0, false, 200 ),
        Limit     ( 0x18D4, 200, false, Limits.values() ),
        ;
        public  final Object[] keys;
        public  final boolean  assigned;
        public  final int      address, count, offset;

        //------------------------------------------------------------------------------------------
        /** ChannelGroup constructor.
         *
         *  @param  address                     the channel group's address .
         *  @param  keys                        the the channel names.
         */
        private ChannelGroup( int address, Object[] keys ){
            this( address, 0, true, keys.length, keys );
        } // Ft817base.ChannelGroup constructor

        //------------------------------------------------------------------------------------------
        /** ChannelGroup constructor.
         *
         *  @param  address                     the channel group's address .
         *  @param  offset                      the channel group's offset.
         *  @param  assigned                    whether the channel group is permanently assigned.
         *  @param  count                       the number of channels.
         */
        private ChannelGroup( int address, int offset, boolean assigned, int count ){
            this( address, offset, assigned, count, null );
        } // Ft817base.ChannelGroup constructor

        //------------------------------------------------------------------------------------------
        /** ChannelGroup constructor.
         *
         *  @param  address                     the channel group's address .
         *  @param  offset                      the channel group's offset.
         *  @param  assigned                    whether the channel group is permanently assigned.
         *  @param  keys                        the the channel names.
         */
        private ChannelGroup( int address, int offset, boolean assigned, Object[] keys ){
            this( address, offset, assigned, keys.length, keys );
        } // Ft817base.ChannelGroup constructor

        //------------------------------------------------------------------------------------------
        /** ChannelGroup constructor.
         *
         *  @param  address                     the channel group's address .
         *  @param  offset                      the channel group's offset.
         *  @param  assigned                    whether the channel group is permanently assigned.
         *  @param  count                       the number of channels.
         *  @param  keys                        the the channel names.
         */
        private ChannelGroup( int address, int offset, boolean assigned, int count, Object[] keys ){
            this.keys     = keys;
            this.assigned = assigned;
            this.address  = address;
            this.offset   = offset;
            this.count    = count;
        } // Ft817base.ChannelGroup constructor

        //------------------------------------------------------------------------------------------
        /** getKeys returns the group's array of keys.
         *
         *  @return Object[]                    - the array of keys.
         */
        public Object[] getKeys(){
            return keys;
        } // Ft817base.ChannelGroup.getAssigned

        //------------------------------------------------------------------------------------------
        /** getAssigned returns whether the group of channels is automatically assigned (visible).
         *
         *  @return boolean                     - whether the group is assigned.
         */
        public boolean getAssigned(){
            return assigned;
        } // Ft817base.ChannelGroup.getAssigned

        //------------------------------------------------------------------------------------------
        /** getAddress returns the group's channel array memory address.
         *
         *  @return int                         - the address.
         */
        public int getAddress(){
            return address;
        } // Ft817base.ChannelGroup.getAddress

        //------------------------------------------------------------------------------------------
        /** getOffset returns the group's channel array memory offset.
         *
         *  @return int                         - the offset.
         */
        public int getOffset(){
            return offset;
        } // Ft817base.ChannelGroup.getOffset

        //------------------------------------------------------------------------------------------
        /** getCount returns the number of channels in the group.
         *
         *  @return int                         - the number of channels.
         */
        public int getCount(){
            return count;
        } // Ft817base.ChannelGroup.getCount

    } // Ft817base.ChannelGroup enum

    public enum Limits      { _P1L, _P1U }
    public enum Temporaries { _Qmb, _MtQmb, _Mtune }
    static public  final int  MEMORY_SIZE  = 0x1908;

    //----------------------------------------------------------------------------------------------
    /** Ft817base constructor.
     *
     *  @param  serialIO                           the SerialIO port to use.
     *  @param  radio                              the radio's memory contents.
     *  @throws IOException                        on an I/O error.
     *  @throws PortInUseException                 if the port is in use.
     *  @throws UnsupportedCommOperationException  on a software error.
     *  @throws TooManyListenersException          on a software error.
     */
    Ft817base( SerialIO serialIO, byte[] radio )        throws  IOException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
        super( serialIO, radio, 26, 0x0450, 0x046A );
    } // Ft817base constructor

    //----------------------------------------------------------------------------------------------
    /** getGroups returns the channel group array.  Overrides Ft___Base.
     * 
     *  @return FtGroup[]                       - the channel groups.
     */
    FtGroup[] getGroups(){
        return ChannelGroup.values();
    } // Ft817base.getGroups

    //----------------------------------------------------------------------------------------------
    /** getGroup returns a channel group by name.  Overrides Ft___Base.
     *
     *  @param  name                            the group name.
     *  @return FtGroup                         - the channel group.
     */
    FtGroup getGroup( String name ){
        return ChannelGroup.valueOf( name );
    } // Ft817base.getGroup

    //----------------------------------------------------------------------------------------------
    /** getChannel returns a radio's Ft___Channel by address.  Overrides Ft___Base.
     *
     *  @param  address                         the channel's address.
     *  @param  state                           the channel's state.
     *  @return Ft___Channel                       - the channel.
     */
    Ft___Channel getChannel( int address, Ft___Channel.States state ){
        return new Ft897channel( radio, address, state );
    } // Ft817base.getChannels

    //----------------------------------------------------------------------------------------------
    /** getCalibration gets the calibration data for the radio.  Overrides Ft___Base.
     *
     *  @return Ft___Calibration                     - the calibration data.
     */
    Ft___Calibration getCalibration(){
        return new Ft817calibration( radio );
    } // Ft817base.getCalibration

    //----------------------------------------------------------------------------------------------
    /** getOptions gets the option data for the radio.  Overrides Ft___Base.
     *
     *  @return Ft___Options                       - the radio's options.
     */
    Ft___Options getOptions(){
        return new Ft817options( radio );
    } // Ft817base.setCalibration

} // class Ft817base
