package net.ae7q.memory;

//==================================================================================================
/** Ft897calibration implements the channel data for an FT-897.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
class Ft897channel extends Ft___Channel {

    //==============================================================================================
    /** Field describes the radio's channels.
     */
    public enum Field implements FtField {
        State            ( 28, 0x03, States.values() ),                                 // MFb:AB
        Skip             (  2, 0x80, FtSettings.Booleans.values() ),                    // MFb:B
        Receive_Freq_Hz  ( 12, 0xFFFFFFFFL, 0, 10, 10000, 47000000 ),                   // Dial
        Band             (  1, 0x07, Bands.values() ),                                  // Band
        Mode             (  0, 0x07, Modes.values() ),                                  // Mode
        Narrow_Band_FM   (  1, 0x08, FtSettings.Booleans.values() ),                    // MFm:C
        Split_Tx         (  1, 0xC0, new String[]{ "SIMPLEX","MINUS","PLUS","SPLIT" } ),// MFd:A +
                                                                                        // MFa:C
        Split_Tx_Freq_Hz ( 16, 0xFFFFFFFFL, 0, 10,     0, 47000000 ),                   // Menu:76
        Split_Tx_Band    (  5, 0x07, Bands.values() ),                                  // Band
        Split_Tx_Mode    (  5, 0x38, Modes.values() ),                                  // Mode
        Clarifier        (  2, 0x40, FtSettings.Booleans.values() ),                    // Clarifier
        Clarifier_Freq_Hz( 10, 0xFFFF,      0, 10, -999, 999 ),                         // Clarifier
        FM_Step          (  3, 0x07, new String[]{  "5kHz","6.25kHz","10kHz","12.5kHz", // Menu:52
                                                   "15kHz",  "20kHz","25kHz",  "50kHz" } ),
        AM_Step          (  3, 0x38, new String[]{"2.5kHz",   "5kHz", "9kHz",           // Menu:06
                                                   "10kHz","12.5kHz","25kHz" } ),
        SSB_Step         (  3, 0xC0, new String[]{  "1kHz", "2.5kHz", "5kHz" } ),       // Menu:82
        RF_Gain          (  2, 0x30, new String[]{ "PRE", "ATT", "IPO" } ),             // MFm:AB
        Tone_Split       (  4, 0x10, FtSettings.Booleans.values() ),                    // Menu:79
        Decode           (  4, 0x03, codings ),                                         // MFe:B
        Encode           (  4, 0x0C, codings ),                                         // MFe:A
        Decode_Tone_CTCSS(  7, 0xFF, ctcssTones ),                                      // Menu:83-R
        Encode_Tone_CTCSS(  6, 0xFF, ctcssTones ),                                      // Menu:83-T
        Decode_DCS_Code  (  9, 0xFF, dcsCodes ),                                        // Menu:33-R
        Encode_DCS_Code  (  8, 0xFF, dcsCodes ),                                        // Menu:33-T
        _unknown0        (  0, 0x20, FtSettings.Booleans.values() ),                    // (lock?)
        _unknown1        (  1, 0x20, FtSettings.Booleans.values() ),                    // (auto)
        Display_Tag      (  0, 0x80, FtSettings.Booleans.values() ),                    // MFb:C
        Default_Tag      (  0, 0x40, FtSettings.Booleans.values() ),                    // Menu:56
        Memory_Tag       ( 20, 0,           0,  8 ),                                    // Menu:56
        ;
        private final Object[] keys;
        private final long     mask, base, min, max;
        private final int      offset, shift, width, scale, length;

        //------------------------------------------------------------------------------------------
        /** Field constructor.
         *
         *  @param  offset                      the field's offset.
         *  @param  mask                        the field's mask.
         */
        private Field( int offset, long mask ){
            this( offset, mask, 0, 1, 0, -1, null );
        } // Ft897channel.Field constructor

        //------------------------------------------------------------------------------------------
        /** Field constructor.
         *
         *  @param  offset                      the field's offset.
         *  @param  mask                        the field's mask.
         *  @param  keys                        the field's keys.
         */
        private Field( int offset, long mask, Object[] keys ){
            this( offset, mask, 0, 1, 0, keys.length - 1, keys );
        } // Ft897channel.Field constructor

        //------------------------------------------------------------------------------------------
        /** Field constructor.
         *
         *  @param  offset                      the field's offset.
         *  @param  mask                        the field's mask.
         *  @param  base                        the field's base.
         *  @param  scale                       the field's scale.
         */
        private Field( int offset, long mask, long base, int scale ){
            this( offset, mask, base, scale, 0, -1, null );
        } // Ft897channel.Field constructor

        //------------------------------------------------------------------------------------------
        /** Field constructor.
         *
         *  @param  offset                      the field's offset.
         *  @param  mask                        the field's mask.
         *  @param  base                        the field's base.
         *  @param  scale                       the field's scale.
         *  @param  min                         the field's minimum value.
         *  @param  max                         the field's maximum value.
         */
        private Field( int offset, long mask, long base, int scale, long min, long max ){
            this( offset, mask, base, scale, min, max, null );
        } // Ft897channel.Field constructor

        //------------------------------------------------------------------------------------------
        /** Field constructor.
         *
         *  @param  offset                      the field's offset.
         *  @param  mask                        the field's mask.
         *  @param  base                        the field's base.
         *  @param  scale                       the field's scale.
         *  @param  min                         the field's minimum value.
         *  @param  max                         the field's maximum value.
         *  @param  keys                        the field's keys.
         */
        private Field( int offset, long mask, long base, int scale, long min, long max,
                       Object[] keys ){
            int shift, width;
            this.offset  = offset;
            this.mask    = mask;
            for ( shift = 0;  mask != 0  && (mask & 1) == 0;  shift++, mask >>>= 1 );
            for ( width = 0;  mask != 0;                      width++, mask >>>= 1 );
            this.shift   = shift;
            this.width   = width;
            this.length  = this.mask != 0 ? (shift + width - 1) / 8 + 1 : scale;
            this.base    = base;
            this.scale   = scale;
            this.min     = min;
            this.max     = max < min ? this.mask >>> shift : max;
            this.keys    = keys;
        } // Ft897channel.Field constructor

        //------------------------------------------------------------------------------------------
        /** get retrieves the field from a buffer.
         *
         *  @param  buffer                      the buffer.
         *  @return Object                      - the field.
         */
        public Object get( byte[] buffer ){
            if ( mask == 0 ){
                return new String( buffer, offset, length );
            } else {
                long value = 0;
                for ( int index = 0;  index < length;  index++  ){
                    value = value << 8 | buffer[ offset + index ] & 0xFF;
                }
                return FtRemote.getValue( (mask & value) >>> shift, base, scale, min, width, keys );
            }
        } // Ft897channel.Field.get

        //------------------------------------------------------------------------------------------
        /** put places a field into a buffer.
         *
         *  @param  buffer                      the buffer.
         *  @param  object                      the field.
         */
        public void put( byte[] buffer, Object object ){
            if ( mask == 0 ){
                byte[] bytes = ((String) object).getBytes();
                for ( int index = 0;  index < length;  index++ ){
                    buffer[ offset + index ] = index < bytes.length ? bytes[ index ] : (byte) ' ';
                }
            } else {
                long mask = this.mask;
                long value = FtRemote.putValue( base, scale, min, max, keys, object ) << shift;
                for ( int index = length;  index-- > 0;  value >>>= 8, mask >>>= 8 ){
                    buffer[ offset + index ] ^= mask & (buffer[ offset + index ] ^ value);
                }
            }
        } // Ft897channel.Field.put

        //------------------------------------------------------------------------------------------
        /** type returns the Type of the field.
         *
         *  @return Type                        - the Type of field.
         */
        public Type type(){
            return mask == 0 ? Type.TAG : Type.OTHER;
        } // Ft897channel.Field.type

    } // Ft897channel.Field enum

    static private final long     serialVersionUID = 2006040900L;
    static         final String[] codings          = { "NONE", "CTCSS", "DCS" };
    static public  final int[]    limits           = { 28, 29 };
           private       byte[]   buffer           = new byte[ limits[ limits.length - 1 ] ];

    //----------------------------------------------------------------------------------------------
    /** Ft897channel constructor.
     *
     *  @param  radio                   the buffer containing the channel data.
     *  @param  address                 the address of the channel data in the buffer.
     *  @param  state                   the channel's state.
     */
    Ft897channel( byte[] radio, int address, States state ){
        for ( int index = 0;  index < limits[ 0 ];  index++ ){
            buffer[ index ] = state != States.UNUSED ? radio[ address++ ] : 0;
        }
        put( Field.State, state.toString() );
    } // Ft897channel constructor

    //----------------------------------------------------------------------------------------------
    /** getFields returns an array of the fields.  Overrides Ft___Settings.
     *
     *  @return FtField[]                       - the array of Fields.
     */
    FtField[] getFields(){
        return Field.values();
    } // Ft897channel.getFields

    //----------------------------------------------------------------------------------------------
    /** getField returns a Field by name.  Overrides Ft___Settings.
     *
     *  @param  name                            the field name.
     *  @return FtField                         - the Field.
     */
    FtField getField( String name ){
        return Field.valueOf( name );
    } // Ft897channel.getField

    //----------------------------------------------------------------------------------------------
    /** set updates the radio buffer from the internal buffer.  Overrides Ft___Channel.
     *
     *  @param  radio                           the radio's buffer.
     *  @param  address                         the address in the radio's buffer.
     *  @return States                          - the channel's state
     */
    States set( byte[] radio, int address ){
        States state = States.valueOf( get( Field.State ).toString() );
        for ( int index = 0;  index < limits[ 0 ];  index++ ){
            radio[ address++ ] = state != States.UNUSED ? buffer[ index ] : -1;
        }
        return state;
    } // Ft897channel.set
    
    //----------------------------------------------------------------------------------------------
    /** getBuffer returns the data buffer.  Overrides Ft___Settings.
     *
     *  @return byte[]                          - the buffer.
     */
    public byte[] getBuffer(){
        return buffer;
    } // Ft897channel.getBuffer

} // class Ft897channel
