package net.ae7q.memory;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.logging.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
// import javax.comm.*;
import gnu.io.*;

import com.ultimeth.util.*;

//==================================================================================================
/** FTremote is a program for setting options in a Yaesu FT-897D.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
class FtRemote extends JApplet {
    static private final long       serialVersionUID = 2010100400L;
    static private final String     TITLE            = "FtRemote for the Yaesu FT-817 and FT-897";
    static private final Collection mapList          = new LinkedList();
    static private final Rectangle  bounds           = new Rectangle( 100, 10, 500, 450 );
    static private       boolean    initialized      = false;

           private final JLabel     statusBar        = new JLabel();

    //==============================================================================================
    /** Radios describes the supported radios.
     */
    public enum Radios {
        Ft897 ( Ft897base.MEMORY_SIZE ){
            Ft___Base getRadio( SerialIO serialIO, byte[] memory )
                                                        throws  IOException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
                return new Ft897base( serialIO, memory );
            } // FtRemote.Radios.getRadio
        },
        Ft817 ( Ft817base.MEMORY_SIZE ){
            Ft___Base getRadio( SerialIO serialIO, byte[] memory )
                                                        throws  IOException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
                return new Ft817base( serialIO, memory );
            } // FtRemote.Radios.getRadio
        },
        ;
        final int memorySize;

        //------------------------------------------------------------------------------------------
        /** Radios constructor.
         *
         *  @param  memorySize                          the radio's memory size.
         */
        private Radios( int memorySize ){
            this.memorySize = memorySize;
        } // FtRemote.Radios constructor

        //------------------------------------------------------------------------------------------
        /** getRadio initializes the radio.
         *
         *  @param  serialIO                            the SerialIO port to use.
         *  @param  memory                              the radio's memory contents.
         *  @return Ft___Base                           - the radio's base object.
         *  @throws IOException                         on an I/O error.
         *  @throws PortInUseException                  if the port is in use.
         *  @throws UnsupportedCommOperationException   on a software error.
         *  @throws TooManyListenersException           on a software error.
         */
        abstract Ft___Base getRadio( SerialIO serialIO, byte[] memory )
                                                        throws  IOException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException;

    } // FtRemote.Radios enum

    //----------------------------------------------------------------------------------------------
    /** Application main entry point.  Create the main window and call common program setup.
     *
     *  @param  args
     */
    static public void main( final String[] args ){
        final JFrame    jFrame    = new JFrame( TITLE );
        final FtDisplay ftDisplay = new FtDisplay( jFrame );

        jFrame.setDefaultCloseOperation( JFrame.DO_NOTHING_ON_CLOSE );
        jFrame.setBounds( ftDisplay.preferences.getInt( "Window.x",      bounds.x ),
                          ftDisplay.preferences.getInt( "Window.y",      bounds.y ),
                          ftDisplay.preferences.getInt( "Window.width",  bounds.width ),
                          ftDisplay.preferences.getInt( "Window.height", bounds.height ) );
        jFrame.setVisible( true );
        jFrame.addWindowListener( new WindowAdapter(){
                public void windowClosing( WindowEvent event ){
                    exit( jFrame );
                }
            } );  // WindowListener implementation

        try {
            // First, initialize settings & GUI
            setup( ftDisplay.log );
            EventQueue.invokeAndWait( new Runnable() {
                    public void run(){
                        jFrame.setBounds( bounds );
                        jFrame.validate();
                    }
                } ); // Runnable implementation
            EventQueue.invokeAndWait( new Runnable() {
                    public void run(){
                        try {
                            ftDisplay.setup();
                        } catch ( Exception exception ){
                            Tools.criticalError( ftDisplay.log, exception );
                            System.exit( 0 );
                        }

                        // Second, set command line options
                        for ( int index = 0;  index < args.length;  index++ ){
                            if ( args[ index ].charAt( 0 ) == '-' 
                                    && args[ index ].length() > 1 ){
                                char option = args[ index ].charAt( 1 );
                                int  mask;
                                switch ( option ){
                                case 'M':
                                case 'U':
                                case 'F':
                                case 'D':
                                    mask = InputEvent.CTRL_MASK | InputEvent.SHIFT_MASK;
                                    break;                              // must be "break"
                                default:                                // must be present
                                    continue;                           // must not be "break"
                                }
                                option = Character.toUpperCase( option );
                                Object value = ftDisplay.keyStrokeMap
                                               .get( KeyStroke.getKeyStroke( option, mask ) );
                                if ( value instanceof JMenuItem ){      // assert
                                    ((JMenuItem) value ).doClick();
                                }
                            }
                        }
                    }
                } ); // Runnable implementation
            ftDisplay.start();
        } catch ( Exception exception ){
            Tools.criticalError( ftDisplay.log, exception );
            System.exit( 0 );
        }
    } // FtRemote.main

    //----------------------------------------------------------------------------------------------
    /** Exit performs normal exit housekeeping tasks.
     *
     *  @param  mainWindow                              the main window Container.
     */
    static public void exit( Container mainWindow ){
        switch ( Tools.showDialog( mainWindow, "Exit",
                              new JOptionPane( "Exit program?",
                                                JOptionPane.QUESTION_MESSAGE,
                                                JOptionPane.OK_CANCEL_OPTION ) ) ){
        case JOptionPane.OK_OPTION:
            System.exit( 0 );
            break;
        }
    }// FtRemote.exit

    //----------------------------------------------------------------------------------------------
    /** Displays text on an Applet status line (JApplet override).
     *
     *  @param  text    the text to display.
     */
    public void showStatus( String text ){
        super.showStatus( text );
        statusBar.setText( "<HTML><B>" + text + "</B></HTML>" );
    }// FtRemote.showStatus

    //----------------------------------------------------------------------------------------------
    /** Applet initialization entry point.  Calls common program setup (JApplet override).
     */
    public void init(){
        Container contentPane = getContentPane();
        contentPane.setLayout( new BoxLayout( contentPane, BoxLayout.Y_AXIS ) );
        contentPane.add( statusBar );
        final FtDisplay ftDisplay = new FtDisplay( this );

        try {
            setup( ftDisplay.log );
            resize( bounds.getSize() );
            ftDisplay.setup();
            new Thread(){
                public void run(){
                    try {
                        ftDisplay.start();
                    } catch ( Exception exception ){
                        Tools.criticalError( ftDisplay.log, exception );
                    }
                }
            }.start();
        } catch ( Exception exception ){
            Tools.criticalError( ftDisplay.log, exception );
        }
    }// FtRemote.init

    //----------------------------------------------------------------------------------------------
    /** Applet start execution entry point (JApplet override).
     */
    public void start(){
    }// FtRemote.start

    //----------------------------------------------------------------------------------------------
    /** Applet stop execution entry point (JApplet override).
     */
    public void stop(){
    }// FtRemote.stop

    //----------------------------------------------------------------------------------------------
    /** Program settings initialization.  Sets up program options.
     *
     *  @param  log                                     the logger to use.
     */
    static private void setup( Logger log ){

        // Get properties and set other options
    }// FtRemote.setup

    //----------------------------------------------------------------------------------------------
    /** getValue decodes a field value for output.
     *
     *  @param  value                                   the value to decode.
     *  @param  base                                    the base to be added to the scaled value.
     *  @param  scale                                   the scale factor.
     *  @param  min                                     the minimum value (used to determine
     *                                                  signedness).
     *  @param  width                                   the field width.
     *  @param  keys                                    the keys (array of Strings or Enums) for
     *                                                  output.
     *  @return Object                                  - the output value (use .toString() if
     *                                                  desired).
     */
    public static Object getValue( long value, long base, int scale, long min, int width,
                                   Object[] keys ){
        if ( keys != null  &&  value < keys.length ){
            return keys instanceof FtEnumPrefix[] ? keys[ (int) value ].toString().substring( 1 )
                                                  : keys[ (int) value ];
        } else {
            long result  = value * scale + base,
                 signBit = 1 << width - 1;
            return min < 0 ? (result ^ signBit) - signBit : result;
        }
    } // FtRemote.getValue

    //----------------------------------------------------------------------------------------------
    /** putValue encodes a field value for storage.
     *
     *  @param  base                                    the base to be added to the scaled value.
     *  @param  scale                                   the scale factor.
     *  @param  min                                     the minimum value.
     *  @param  max                                     the maximum value.
     *  @param  keys                                    the keys (array of Strings or Enums)
     *                                                  for matching.
     *  @param  object                                  the value to encode.
     *  @return long                                    - the field value.
     *  @throws IllegalArgumentException                on an invalid value.
     */
    public static long putValue( long base, int scale, long min, long max, Object[] keys,
                                 Object object )        throws  IllegalArgumentException {
        String string = object.toString();
        if ( keys != null ){
            if ( keys instanceof Enum[] ){
                return getEnum( keys, (keys instanceof FtEnumPrefix[] ? "_" : "") + string )
                       .ordinal();
            }
            for ( int value = 0;  value < keys.length;  value++ ){
                if ( string.equals( keys[ value ] ) ){
                    return value;
                }
            }
        }
        if ( object instanceof Boolean ){
            return (Boolean) object ? 1 : 0;
        }
        long value = (Long.parseLong( string ) - base) / scale;
        if ( value >= min  &&  value <= max ){
            return value;
        }
        throw new IllegalArgumentException( string );
    } // FtRemote.putValue

    //----------------------------------------------------------------------------------------------
    /** getEnum matches an Enum to a string
     *
     *  @param  enums                                   the list of Enum values.
     *  @param  name                                    the name to match.
     *  @return Enum                                    - the matching Enum.
     *  @throws IllegalArgumentException                on an invalid value.
     */
    @SuppressWarnings( "unchecked" )
    public static Enum getEnum( Object[] enums, String name ){
         return Enum.valueOf( ((Enum) enums[ 0 ]).getClass(), name );  // [unchecked]
    } // FtRemote.getEnum

} // class FtRemote

//==================================================================================================
/** FtEnumPrefix flags an Enum has having names with leading underscores that should be removed
 *  for display.
 */
interface FtEnumPrefix {
} // interface FtEnumPrefix

//==================================================================================================
/** FtField characterizes Enums describing fields.
 */
interface FtField {
    public enum Type { TAG, MORSE, SIGNED, OTHER }

    //----------------------------------------------------------------------------------------------
    /** get retrieves the field from a buffer.
     *
     *  @param  buffer                                  the buffer.
     *  @return Object                                  - the field.
     */
    Object get( byte[] buffer );

    //----------------------------------------------------------------------------------------------
    /** put places a field into a buffer.
     *
     *  @param  buffer                                  the buffer.
     *  @param  object                                  the field.
     */
    void put( byte[] buffer, Object object );

    //----------------------------------------------------------------------------------------------
    /** type returns the Type of the field.
     *
     *  @return Type                                    - the Type of the field.
     */
     Type type();

} // interface FtField

//==================================================================================================
/** FtGroup characterizes Enums describing channel groups.
 */
interface FtGroup {

    //----------------------------------------------------------------------------------------------
    /** getKeys returns the group's array of keys.
     *
     *  @return Object[]                                - the array of keys.
     */
    Object[] getKeys();

    //----------------------------------------------------------------------------------------------
    /** getAssigned returns whether the group of channels is automatically assigned (visible).
     *
     *  @return boolean                                 - whether the group is assigned.
     */
    boolean getAssigned();

    //----------------------------------------------------------------------------------------------
    /** getAddress returns the group's channel array memory address.
     *
     *  @return int                                     - the address.
     */
    int getAddress();

    //----------------------------------------------------------------------------------------------
    /** getOffset returns the group's channel array memory offset.
     *
     *  @return int                                     - the offset.
     */
    int getOffset();

    //----------------------------------------------------------------------------------------------
    /** getCount returns the number of channels in the group.
     *
     *  @return int                                     - the number of channels.
     */
    int getCount();

} // interface FtGroup

//==================================================================================================
/** FtChannelGroup characterizes a group of channels.
 */
class FtChannelGroup extends AbstractTableModel {

    //==============================================================================================
    /** Column names and labels.
     */
    private enum Columns {
        ID              ( "Channel ID" ),
        Receive_Freq_Hz ( "Rx Frequency" ),
        Split_Tx_Freq_Hz( "Tx Freq/Split" ),
        Memory_Tag      ( "Label" ),
        ;
        public  final String label;

        //------------------------------------------------------------------------------------------
        /** Bands constructor.
         *
         *  @param  label                       the label to display.
         */
        Columns( String label ){
            this.label = label;
        } // FtChannelGroup.Columns constructor
    } // FtChannelGroup.Columns enum

    static private final long           serialVersionUID = 2006040900L;
           private final FtGroup        group;
           private final Ft___Channel[] channels;

    //----------------------------------------------------------------------------------------------
    /** FtChannels constructor.
     *
     *  @param  group                           the channel group.
     *  @param  radio                           the radio base.
     */
    public FtChannelGroup( FtGroup group, Ft___Base radio ){
        this.group = group;
        channels   = radio.getChannels( group );
    }// FtChannelGroup constructor

    //----------------------------------------------------------------------------------------------
    /** Gets the number of TableRows (AbstractTableModel override).
     *
     *  @return int             - the number of TableRows.
     */
    public int getRowCount(){
        return channels.length;
    }// FtChannelGroup.getRowCount

    //----------------------------------------------------------------------------------------------
    /** Gets the number of columns (AbstractTableModel override).
     *
     *  @return int             - the number of columns.
     */
    public int getColumnCount(){
        return Columns.values().length;
    }// FtChannelGroup.getColumnCount

    //----------------------------------------------------------------------------------------------
    /** Gets a column name (AbstractTableModel override).
     *
     *  @param  column          the column index.
     *  @return String          - the column name.
     */
    public String getColumnName( int column ){
        return Columns.values()[ column ].toString();
    }// FtChannelGroup.getColumnName

    //----------------------------------------------------------------------------------------------
    /** Gets a column Object (AbstractTableModel override).
     *
     *  @param  row             the subscript of the Object's row.
     *  @param  column          the subscript of the Object's column.
     *  @return Object          - the selected Object.
     */
    public Object getValueAt( int row, int column ){
        if ( column > 0 ){
            return channels[ row ].get( Columns.values()[ column ].toString() );
        }
        Object[] keys = group.getKeys();
        if ( keys == null ){
            return String.valueOf( row + 1001 ).substring( 1 );
        }
        Object key = keys[ row ];
        return key instanceof FtEnumPrefix ? key.toString().substring( 1 ) : key;
    }// FtChannelGroup.getValueAt

} // class FtChannelGroup

//==================================================================================================
/** FtSettings characterizes a group of radio settings.
 */
abstract class FtSettings extends AbstractTableModel {

    public enum Booleans { OFF, ON }

           public  final String[] labels;

    //----------------------------------------------------------------------------------------------
    /** FtSettings constructor.
     *
     *  @param  nameLabel                       the name column label.
     *  @param  valueLabel                      the value column label.
     */
    public FtSettings( String nameLabel, String valueLabel ){
        labels = new String[] { nameLabel, valueLabel };
    }// FtSettings constructor

    //----------------------------------------------------------------------------------------------
    /** Gets the number of TableRows (AbstractTableModel override).
     *
     *  @return int             - the number of TableRows.
     */
    public int getRowCount(){
        return getFields().length;
    }// FtSettings.getRowCount

    //----------------------------------------------------------------------------------------------
    /** Gets the number of columns (AbstractTableModel override).
     *
     *  @return int             - the number of columns.
     */
    public int getColumnCount(){
        return labels.length;
    }// FtSettings.getColumnCount

    //----------------------------------------------------------------------------------------------
    /** Gets a column name (AbstractTableModel override).
     *
     *  @param  column          the column index.
     *  @return String          - the column name.
     */
    public String getColumnName( int column ){
        return labels[ column ];
    }// FtSettings.getColumnName

    //----------------------------------------------------------------------------------------------
    /** Gets a column Object (AbstractTableModel override).
     *
     *  @param  row             the subscript of the Object's row.
     *  @param  column          the subscript of the Object's column.
     *  @return Object          - the selected Object.
     */
    public Object getValueAt( int row, int column ){
        FtField field = getFields()[ row ];
        return column == 0 ? field instanceof FtEnumPrefix ? field.toString().substring( 1 )
                                                           : field
                           : get( field );
    }// FtSettings.getValueAt

    //----------------------------------------------------------------------------------------------
    /** getBuffer returns the data buffer.
     *
     *  @return byte[]                                  - the buffer.
     */
    abstract byte[] getBuffer();

    //----------------------------------------------------------------------------------------------
    /** getFields returns an array of the fields.
     *
     *  @return FtField[]                               - the array of Fields.
     */
    abstract FtField[] getFields();

    //----------------------------------------------------------------------------------------------
    /** getField returns a Field by name.
     *
     *  @param  name                                    the field name.
     *  @return FtField                                 - the Field.
     */
    abstract FtField getField( String name );

    //----------------------------------------------------------------------------------------------
    /** get returns a field value.
     *
     *  @param  name                                    the field name.
     *  @return Object                                  - the value.
     */
    Object get( String name ){
        return get( getField( name ) );
    } // FtSettings.get

    //----------------------------------------------------------------------------------------------
    /** get returns a field value.
     *
     *  @param  field                                   the Field.
     *  @return Object                                  - the value.
     */
    Object get( FtField field ){
        return field.get( getBuffer() );
    } // FtSettings.get
    
    //----------------------------------------------------------------------------------------------
    /** put inserts a value into a Field
     *
     *  @param  field                                   the Field.
     *  @param  object                                  the value.
     */
    void put( FtField field, Object object ){
        field.put( getBuffer(), object );
    } // FtSettings.put

    //----------------------------------------------------------------------------------------------
    /** toString creates a hexadecimal dump string of the data.
     *
     *  @return String                                  - the dump of the data.
     */
    public String toString(){
        return SerialIO.dump( getBuffer(), 0, getBuffer().length, 16, true );
    } // FtSettings.toString

} // class FtSettings

//==================================================================================================
/** Ft___Calibration characterizes a radio's calibration settings.
 */
abstract class Ft___Calibration extends FtSettings {

    //----------------------------------------------------------------------------------------------
    /** Ft___Calibration constructor.
     */
    public Ft___Calibration(){
        super( "Calibration option", "Calibration value" );
    }// Ft___Calibration constructor

    //----------------------------------------------------------------------------------------------
    /** set updates the radio buffer from the internal buffer.
     *
     *  @param  radio                                   the radio's buffer.
     */
    abstract void set( byte[] radio );

} // class Ft___Calibration

//==================================================================================================
/** Ft___Channel characterizes a radio's channel.
 */
abstract class Ft___Channel extends FtSettings {

    public enum States { UNUSED, ASSIGNED, VISIBLE }
    public enum Modes  { LSB, USB, CW, CWR, AM, FM, DIG, PKT }
    static         final String[] ctcssTones
        = {"67.0",  "69.3",  "71.9",  "74.4",  "77.0",  "79.7",  "82.5",  "85.4",  "88.5",  "91.5",
           "94.8",  "97.4", "100.0", "103.5", "107.2", "110.9", "114.8", "118.8", "123.0", "127.3",
          "131.8", "136.5", "141.3", "146.2", "151.4", "156.7", "159.8", "162.2", "165.5", "167.9",
          "171.3", "173.8", "177.3", "179.9", "183.5", "186.2", "189.9", "192.8", "196.6", "199.5",
          "203.5", "206.5", "210.7", "218.1", "225.7", "229.1", "233.6", "241.8", "250.3", "254.1"};
    static         final String[] dcsCodes
        = { "023", "025", "026", "031", "032", "036", "043", "047", "051", "053",
            "054", "065", "071", "072", "073", "074", "114", "115", "116", "122",
            "125", "131", "132", "134", "143", "145", "152", "155", "156", "162",
            "165", "172", "174", "205", "212", "223", "225", "226", "243", "244",
            "245", "246", "251", "252", "255", "261", "263", "265", "266", "271",
            "274", "306", "311", "315", "325", "331", "332", "343", "346", "351",
            "356", "364", "365", "371", "411", "412", "413", "423", "431", "432",
            "445", "446", "452", "454", "455", "462", "464", "465", "466", "503",
            "506", "516", "523", "526", "532", "546", "565", "606", "612", "624",
            "627", "631", "632", "654", "662", "664", "703", "712", "723", "731",
            "732", "734", "743", "754" };

    //==============================================================================================
    /** Bands describes the bands used by channels.
     */
    public enum Bands implements FtEnumPrefix {
        _HF  (    100,  33000 ),
        _6M  (  33000,  56000 ),
        _FM  (  76000, 108000 ),
        _AIR ( 118000, 137000 ),
        _VHF ( 137000, 164000 ),
        _UHF ( 420000, 470000 ),
        ;
        int lower, upper;

        //------------------------------------------------------------------------------------------
        /** Bands constructor.
         *
         *  @param  lower                               the lower frequency limit in kHz.
         *  @param  upper                               the upper frequency limit in kHz.
         */
        Bands( int lower, int upper ){
            this.lower = lower;
            this.upper = upper;
        } // Ft___Channel.Bands constructor
    } // Ft___Channel.Bands enum

    //----------------------------------------------------------------------------------------------
    /** Ft___Channel constructor.
     */
    public Ft___Channel(){
        super( "Channel setting", "Option value" );
    }// Ft___Channel constructor

    //----------------------------------------------------------------------------------------------
    /** set updates the radio buffer from the internal buffer.
     *
     *  @param  radio                                   the radio's buffer.
     *  @param  offset                                  the offset into the radio's buffer.
     *  @return States                                  - the channel's state
     */
    abstract States set( byte[] radio, int offset );

} // class Ft___Channel

//==================================================================================================
/** Ft___Options characterizes a radio's options.
 */
abstract class Ft___Options extends FtSettings {

    //==============================================================================================
    /** Bands describes the bands used by channels.
     */
    public enum HomeBands implements FtEnumPrefix {
        _HF  (    100,  33000 ),
        _6M  (  33000,  56000 ),
        _VHF ( 137000, 164000 ),
        _UHF ( 420000, 470000 ),
        ;
        int lower, upper;

        //------------------------------------------------------------------------------------------
        /** HomeBands constructor.
         *
         *  @param  lower                               the lower frequency limit in kHz.
         *  @param  upper                               the upper frequency limit in kHz.
         */
        HomeBands( int lower, int upper ){
            this.lower = lower;
            this.upper = upper;
        } // Ft___Options.HomeBands constructor
    } // Ft___Options.HomeBands enum

    //----------------------------------------------------------------------------------------------
    /** Ft___Options constructor.
     */
    public Ft___Options(){
        super( "Setting name", "Setting value" );
    }// Ft___Options constructor

    //----------------------------------------------------------------------------------------------
    /** set updates the radio buffer from the internal buffer.
     *
     *  @param  radio                                   the radio's buffer.
     */
    abstract void set( byte[] radio );

} // class Ft___Options

//==================================================================================================
/** Ft___Base characterizes a radio.
 */
abstract class Ft___Base {

    final int      channelSize,
                   memoryVisible,
                   memoryAssigned;
    final SerialIO serialIO;
    final byte[]   radio;

    //----------------------------------------------------------------------------------------------
    /** Ft___Base constructor.
     *
     *  @param  serialIO                                the SerialIO port to use.
     *  @param  radio                                   the radio's memory contents.
     *  @param  channelSize                             the size of a memory channel.
     *  @param  memoryVisible                           the location of the visible memory bit arrar.
     *  @param  memoryAssigned                          the location of the assigned memory bit array.
     *  @throws IOException                             on an I/O error.
     *  @throws PortInUseException                      if the port is in use.
     *  @throws UnsupportedCommOperationException       on a software error.
     *  @throws TooManyListenersException               on a software error.
     */
    Ft___Base( SerialIO serialIO, byte[] radio,
               int channelSize, int memoryVisible, int memoryAssigned )
                                                        throws  IOException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
        this.serialIO       = serialIO;
        this.radio          = radio;
        this.channelSize    = channelSize;
        this.memoryVisible  = memoryVisible;
        this.memoryAssigned = memoryAssigned;
    } // Ft___Base constructor

    //----------------------------------------------------------------------------------------------
    /** getChannel returns a radio's Ft___Channel by address.
     *
     *  @param  address                                 the channel's address.
     *  @param  state                                   the channel's state.
     *  @return Ft___Channel                            - the channel.
     */
    abstract Ft___Channel getChannel( int address, Ft___Channel.States state );

    //----------------------------------------------------------------------------------------------
    /** getGroups returns the channel group array.
     * 
     *  @return FtGroup[]                               - the channel groups.
     */
    abstract FtGroup[] getGroups();

    //----------------------------------------------------------------------------------------------
    /** getGroup returns a channel group by name.
     *
     *  @param  name                                    the group name.
     *  @return FtGroup                                 - the channel group.
     */
    abstract FtGroup getGroup( String name );

    //----------------------------------------------------------------------------------------------
    /** getChannels gets a group's array of channels.
     *
     *  @param  group                                   the group.
     *  @return Ft___Channel[]                          - the array of channels.
     */
    Ft___Channel[] getChannels( FtGroup group ){
        Ft___Channel[] result = new Ft___Channel[ group.getCount() ];
        int address = group.getAddress();
        for ( int index = 0;  index < result.length;  index++, address += channelSize ){
            Ft___Channel.States state = Ft___Channel.States.VISIBLE;
            if ( ! group.getAssigned() ){
                if ( getBit( radio, memoryAssigned, group.getOffset() + index ) ){
                    state = getBit( radio, memoryVisible, group.getOffset() + index )
                          ? Ft___Channel.States.VISIBLE : Ft___Channel.States.ASSIGNED;
                } else {
                    state = Ft___Channel.States.UNUSED;
                }
            }
            result[ index ] = getChannel( address, state );
        }
        return result;
    } // Ft___Base.getChannels

    //----------------------------------------------------------------------------------------------
    /** setChannels updates the radio's group of channels.
     *
     *  @param  group                                   the group.
     *  @param  channels                                the array of channels with updated data.
     *  @return int                                     - the number of bytes changed.
     *  @throws IOException                             on a I/O error.
     */
    int setChannels( FtGroup group, Ft___Channel[] channels )
                                                        throws  IOException,
                                                                InterruptedException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
        byte[] updates = radio.clone();
        int    address = group.getAddress();
        for ( int index = 0;  index < group.getCount();  index++, address += channelSize ){
            Ft___Channel.States state  = channels[ index ].set( updates, address );
            if ( ! group.getAssigned() ){
                int offset = group.getOffset() + index;
                putBit( updates, memoryAssigned, offset, state != Ft___Channel.States.UNUSED );
                putBit( updates, memoryVisible,  offset, state == Ft___Channel.States.VISIBLE );
            }
        }
        return serialIO.update( radio, 0, 0, updates, 0, radio.length );
    } // Ft___Base.setChannels

    //----------------------------------------------------------------------------------------------
    /** getCalibration gets the calibration data for the radio.
     *
     *  @return Ft___Calibration                        - the calibration data.
     */
    abstract Ft___Calibration getCalibration();

    //----------------------------------------------------------------------------------------------
    /** setCalibration updates the calibration data in the radio.
     *
     *  @param  calibration                             the calibration data.
     *  @return int                                     - the number of bytes changed.
     *  @throws IOException                             on an I/O error.
     *  @throws InterruptedException                    on a timeout.
     *  @throws PortInUseException                      if the port is in use.
     *  @throws UnsupportedCommOperationException       on a software error.
     *  @throws TooManyListenersException               on a software error.
     */
    int setCalibration( Ft___Calibration calibration )  throws  IOException,
                                                                InterruptedException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
        byte[] updates = radio.clone();
        calibration.set( updates );
        return serialIO.update( radio, 0, 0, updates, 0, radio.length );
    } // Ft___Base.getCalibration

    //----------------------------------------------------------------------------------------------
    /** getOptions gets the option data for the radio.
     *
     *  @return Ft___Options                            - the radio's options.
     */
    abstract Ft___Options getOptions();

    //----------------------------------------------------------------------------------------------
    /** setOptions updates the option data in the radio.
     *
     *  @param  options                                 the radio's options.
     *  @return int                                     - the number of bytes changed.
     *  @throws IOException                             on an I/O error.
     *  @throws InterruptedException                    on a timeout.
     *  @throws PortInUseException                      if the port is in use.
     *  @throws UnsupportedCommOperationException       on a software error.
     *  @throws TooManyListenersException               on a software error.
     */
    int setOptions( Ft___Options options )              throws  IOException,
                                                                InterruptedException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
        byte[] updates = radio.clone();
        options.set( updates );
        return serialIO.update( radio, 0, 0, updates, 0, radio.length );
    } // Ft___Base.setOptions

    //----------------------------------------------------------------------------------------------
    /** getBit gets a bit from a bit array.
     *
     *  @param  buffer                                  the buffer to use.
     *  @param  address                                 the address of the first bit.
     *  @param  index                                   the bit index.
     *  @return boolean                                 - the bit value.
     */
    boolean getBit( byte[] buffer, int address, int index ){
        return (buffer[ address + index / 8 ] & 1 << index % 8) != 0;
    } // Ft___Base.get

    //----------------------------------------------------------------------------------------------
    /** putBit sets a bit into a bit array.
     *
     *  @param  buffer                                  the buffer to use.
     *  @param  address                                 the address of the first bit.
     *  @param  index                                   the bit index.
     *  @param  value                                   the bit value.
     */
    void putBit( byte[] buffer, int address, int index, boolean value ){
        int temp = (buffer[ address + index / 8 ] ^ (value ? -1 : 0)) & 1 << index % 8;
        buffer[ address + index / 8 ] ^= temp;
    } // Ft___Base.put

    //----------------------------------------------------------------------------------------------
    /** dump creates a hexadecimal dump string of the selected portion of a radio's buffer.
     *
     *  @param  offset                                  the offset to bump.
     *  @param  limit                                   the offset limit to dump.
     *  @param  width                                   the number of bytes displayed on a line.
     *  @return String                                  - the dump.
     */
    String dump( int offset, int limit, int width ){
        return SerialIO.dump( radio, offset, limit - offset, width, true );
    } // Ft___Base.dump

} // class Ft___Base
