package net.ae7q.memory;

import java.io.*;
import java.util.*;
// import javax.comm.*;
import gnu.io.*;

import com.ultimeth.util.*;

//==================================================================================================
/** Ic7000 implements the serial IO routines for communicating with the radio.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
class Ic7000 implements SerialPortEventListener {
    static private final int          MAX_RADIO  = 0x2000;
           private       SerialPort   serialPort = null;
           private       InputStream  iStream    = null;
           private       OutputStream oStream    = null;

// IC-7000 memory read/write:
// Command: 1A 00 01-5: bank#] [0001-105: mem#]
//    Data: [0-1: split; 0?]
//     Rx:  [89 67 45 23 01: 123.456.789Hz frequency] [00-8: mode] [01-3: filter] [0-2: DUP; 0-3: tone type]
//          [00 12 34: 123.4 Tx CTCSS tone] [00 12 34: 123.4 Rx CTCSS tone] [0-1, 0-1: DTCS polarity] [12 34: 1234 DTCS code]
//     Spl: (same)

    //----------------------------------------------------------------------------------------------
    /** Application main entry point.  Dump a radio's memory.
     *
     *  @param  args                            an array of command line arguments:<OL>
     *                                          <LI> Serial port name (eg, COM1)
     *                                          </OL>
     */
    @SuppressWarnings( "unchecked" )
    static public void main( String[] args ){
        try {
            TreeMap <String, CommPortIdentifier> map = new TreeMap <String, CommPortIdentifier> ();
            for ( CommPortIdentifier commPortId : new ToIterable <CommPortIdentifier>(
                                                      CommPortIdentifier.getPortIdentifiers() ) ){  // [unchecked]
                if ( commPortId.getPortType() == CommPortIdentifier.PORT_SERIAL ){
                    if ( commPortId.getName().equalsIgnoreCase( args[ 0 ] ) ){
                        System.err.println( "Enter IC-7000 commands ..." );
                        Ic7000 ic7000 = new Ic7000( commPortId, 19200, 100 );
                        return;
                    } else {
                        map.put( commPortId.getName(), commPortId );
                    }
                }
            }
            System.err.println( "*** Port " + args[ 0 ] + " not found;  available ports are:" );
            for ( CommPortIdentifier commPortId : map.values() ){
                System.err.println( "  " + commPortId.getName() );
            }
        } catch ( Exception exception ){
            exception.printStackTrace();
        } finally {
            System.exit( 0 );  // terminates the JRE serial IO thread
        }
    } // Ic7000.main

    //----------------------------------------------------------------------------------------------
    /** Ic7000 constructor.
     *
     *  @param  serialPortId                    the serial port to use.
     *  @param  baudRate                        the BAUD rate to use.
     *  @param  timeOut                         the timeout in millseconds to use.
     */
    Ic7000( CommPortIdentifier serialPortId, int baudRate, int timeOut ){
        BufferedReader reader  = new BufferedReader( new InputStreamReader( System.in ) );
        byte[]         command = new byte[ 100 ];
        String         line;
        String[]       tokens;

        command[ 0 ]      = (byte) 0xFE;
        command[ 1 ]      = (byte) 0xFE;
        command[ 2 ]      = (byte) 0x70;
        command[ 3 ]      = (byte) 0xE0;
        try {
            serialPort = (SerialPort) serialPortId.open( "Ic7000", timeOut );
            iStream = serialPort.getInputStream();
            oStream = serialPort.getOutputStream();
            serialPort.setSerialPortParams( baudRate, SerialPort.DATABITS_8, SerialPort.STOPBITS_1,
                                                      SerialPort.PARITY_NONE );
            serialPort.setFlowControlMode( SerialPort.FLOWCONTROL_NONE );
            serialPort.setInputBufferSize( MAX_RADIO );
//          serialPort.enableReceiveTimeout( timeOut );
            serialPort.addEventListener( this );
            serialPort.notifyOnDataAvailable( true );
//          for ( int skip;  (skip = iStream.available()) > 0;  iStream.skip( skip ) );
            while ( (line = reader.readLine()) != null ){
                tokens = line.split( " " );
                if ( tokens == null  ||  tokens[ 0 ].equalsIgnoreCase( "q" ) ){
                    break;
                }
                try {
                    int length = 4;
                    for ( String token : tokens ){
                        if ( token.length() > 0 ){
                            command[ length++ ] = (byte) Short.parseShort( token, 16 );
                        }
                    }
                    command[ length++ ] = (byte) 0xFD;
                    oStream.write( command, 0, length );
                } catch ( Exception exception ){
                    System.out.println( exception.toString() );
                }
            }
        } catch ( Exception exception ){
            exception.printStackTrace();
        }
    } // Ic7000 constructor

    //----------------------------------------------------------------------------------------------
    /** serialEvent (inherited from SerialPortEventListener).
     *
     *  @param  event                           the event (unused).
     */
    public synchronized void serialEvent( SerialPortEvent event ){
        try {
            int data;
            while ( (data = iStream.read()) >= 0 ){
                System.err.print( Integer.toHexString( data ) + (data == 0xFD ? "\n" : " ") );
            }
        } catch ( Exception exception ){
            exception.printStackTrace();
        }
//      notify();
    } // Ic7000.serialEvent

} // class Ic7000
