package net.ae7q.memory;

import java.io.*;
import java.util.*;
// import javax.comm.*;
import gnu.io.*;

import com.ultimeth.util.*;

//==================================================================================================
/** SerialIO implements the serial IO routines for communicating with the radio.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
class SerialIO implements SerialPortEventListener {
    static private final int                MAX_RADIO    = 0x2000;
           private final CommPortIdentifier serialPortId;
           private       SerialPort         serialPort   = null;
           private       InputStream        iStream;
           private       OutputStream       oStream;
           private final int                baudRate;
           private       int                timeOut;

    //----------------------------------------------------------------------------------------------
    /** Application main entry point.  Dump a radio's memory.
     *
     *  @param  args                            an array of command line arguments:<OL>
     *                                          <LI> Serial port name (eg, COM1)
     *                                          </OL>
     */
    @SuppressWarnings( "unchecked" )
    static public void main( String[] args ){
        try {
            TreeMap <String, CommPortIdentifier> map = new TreeMap <String, CommPortIdentifier> ();
            for ( CommPortIdentifier commPortId
                  : new ToIterable <CommPortIdentifier>(
                        CommPortIdentifier.getPortIdentifiers() ) ){  // [unchecked]
                if ( commPortId.getPortType() == CommPortIdentifier.PORT_SERIAL ){
                    if ( commPortId.getName().equalsIgnoreCase( args[ 0 ] ) ){
                        System.err.println( "Loading radio values ..." );
                        SerialIO serialIO = new SerialIO( commPortId, 38400, 100 );
                        byte[] radio = serialIO.getMemory();
                        System.err.format( "Length: %X", radio.length );
                        System.out.println( dump( radio, 0, radio.length, 16, true ) );
                        return;
                    } else {
                        map.put( commPortId.getName(), commPortId );
                    }
                }
            }
            System.err.println( "*** Port " + args[ 0 ] + " not found;  available ports are:" );
            for ( CommPortIdentifier commPortId : map.values() ){
                System.err.println( "  " + commPortId.getName() );
            }
        } catch ( Exception exception ){
            exception.printStackTrace();
        } finally {
            System.exit( 0 );  // terminates the JRE serial IO thread
        }
    } // SerialIO.main

    //----------------------------------------------------------------------------------------------
    /** SerialIO constructor.
     *
     *  @param  serialPortId                    the serial port to use.
     *  @param  baudRate                        the BAUD rate to use.
     *  @param  timeOut                         the timeout in millseconds to use.
     */
    SerialIO( CommPortIdentifier serialPortId, int baudRate, int timeOut ){
        this.serialPortId = serialPortId;
        this.baudRate     = baudRate;
        this.timeOut      = timeOut;
    } // SerialIO constructor

    //----------------------------------------------------------------------------------------------
    /** serialEvent (inherited from SerialPortEventListener).
     *
     *  @param  event                           the event (unused).
     */
    public synchronized void serialEvent( SerialPortEvent event ){
        notify();
    } // SerialIO.serialEvent

    //----------------------------------------------------------------------------------------------
    /** update updates the radio with only the bytes that have changed.
     *
     *  @param  oldBuffer                          the buffer with the old data.
     *  @param  oldOffset                          the offset into the oldBuffer.
     *  @param  address                            the memory address in the radio.
     *  @param  newBuffer                          the buffer with the new data.
     *  @param  newOffset                          the offset into the newBuffer.
     *  @param  length                             the number of bytes to update.
     *  @return int                                - the number of bytes changed.
     *  @throws IOException                        on an I/O error.
     *  @throws InterruptedException               on a timeout.
     *  @throws PortInUseException                 if the port is in use.
     *  @throws UnsupportedCommOperationException  on a software error.
     *  @throws TooManyListenersException          on a software error.
     */
    synchronized int update( byte[] oldBuffer, int oldOffset, int address,
                             byte[] newBuffer, int newOffset, int length )
                                                        throws  IOException,
                                                                InterruptedException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
        byte[] command = new byte[]{ 0, 0, 0, 0, (byte) 0xBC };
        int    count   = 0;
        try {
            for ( ;  length-- > 0;  oldOffset++, newOffset++, address++ ){
                if ( oldBuffer[ oldOffset ] != newBuffer[ newOffset ] ){
                    if ( length-- > 0 ){
                        command[ 2 ] = newBuffer[ newOffset++ ];
                        oldBuffer[ oldOffset++ ] = command[ 2 ];
                    } else {
                        command[ 2 ] = oldBuffer[ oldOffset - 1 ];
                        address--;
                    }
                    count += oldBuffer[ oldOffset ] != newBuffer[ newOffset ] ? 2 : 1;
                    command[ 3 ] = newBuffer[ newOffset ];
                    oldBuffer[ oldOffset ] = command[ 3 ];
                    command[ 0 ] = (byte) (address >> 8);
                    command[ 1 ] = (byte)  address++;
                    if ( serialPort == null ){
                        open();
                    }
                    oStream.write( command );
                    wait( timeOut );
                    if ( iStream.read() < 0 ){
                        throw new IOException( "Timeout writing to radio" );
                    }
                }
            }
        } finally {
            close();
        }
        return count;
    } // SerialIO.update

    //----------------------------------------------------------------------------------------------
    /** getMemory returns the memory for a radio.
     *
     *  @return byte[]                             - the memory bytes read.
     *  @throws IOException                        on an I/O error.
     *  @throws PortInUseException                 if the port is in use.
     *  @throws UnsupportedCommOperationException  on a software error.
     *  @throws TooManyListenersException          on a software error.
     */
    byte[] getMemory()                                  throws  IOException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
        byte[] buffer = new byte[ MAX_RADIO ];
        int offset = 0;
        open();
        for ( int length = buffer.length / 2;  length > 1;  length /= 2 ){
            if ( get( offset, length, buffer, offset ) == length ){
                offset += length;
            }
        }
        if ( get( offset - 1, 2, buffer, offset ) > 1
             &&  buffer[ offset - 1 ] == buffer[ offset ] ){
            buffer[ offset ] = buffer[ offset + 1 ];
            offset++;
        }
        close();
        byte[] result = new byte[ offset ];
        for ( int index = 0;  index < offset;  index++ ){
            result[ index ] = buffer[ index ];
        }
        return result;
    } // SerialIO.getMemory

    //----------------------------------------------------------------------------------------------
    /** get reads memory locations in the radio.
     *
     *  @param  address                            the address in the radio.
     *  @param  length                             the number of bytes to read.
     *  @param  buffer                             the buffer to read memory into.
     *  @param  offset                             the offset in the buffer.
     *  @return int                                - the number bytes read.
     *  @throws IOException                        on an I/O error.
     */
    private int get( int address, int length, byte[] buffer, int offset )
                                                        throws  IOException {
        byte[] command = new byte[]{ 0, 0, 0, 0, (byte) 0xBB };
        int    data;
        for ( int count = length / 2;  count-- > 0;  address += 2 ){
            command[ 0 ] = (byte) (address >> 8);
            command[ 1 ] = (byte)  address;
            oStream.write( command );
        }
        for ( int count = 0;  count < length;  count++ ){
            if ( (data = iStream.read()) < 0 ){
                return count;
            }
            buffer[ offset++ ] = (byte) data;
        }
        return length;
    } // SerialIO.get

    //----------------------------------------------------------------------------------------------
    /** open opens the serial port for reading and/or writing.
     *
     *  @throws IOException                        on an I/O error.
     *  @throws PortInUseException                 if the port is in use.
     *  @throws UnsupportedCommOperationException  on a software error.
     *  @throws TooManyListenersException          on a software error.
     */
    private void open()                                 throws  IOException,
                                                                PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException {
        serialPort = (SerialPort) serialPortId.open( "FTremote", timeOut );
        iStream = serialPort.getInputStream();
        oStream = serialPort.getOutputStream();
        serialPort.setSerialPortParams( baudRate, SerialPort.DATABITS_8, SerialPort.STOPBITS_1,
                                                  SerialPort.PARITY_NONE );
        serialPort.setFlowControlMode( SerialPort.FLOWCONTROL_NONE );
        serialPort.setInputBufferSize( MAX_RADIO );
        serialPort.enableReceiveTimeout( timeOut );
        serialPort.addEventListener( this );
        serialPort.notifyOnDataAvailable( true );
        for ( int skip;  (skip = iStream.available()) > 0;  iStream.skip( skip ) );
    } // SerialIO.open

    //----------------------------------------------------------------------------------------------
    /** close closes the serial port.
     */
    private void close(){
        if ( serialPort != null ){
            serialPort.close();
            serialPort = null;
        }
    } // SerialIO.close

    //----------------------------------------------------------------------------------------------
    /** dump creates a hexadecimal dump string of the selected portion of a byte array.
     *
     *  @param  buffer                          the buffer to dump.
     *  @param  offset                          the offset to bump.
     *  @param  size                            the size of the segment to dump.
     *  @param  width                           the number of bytes displayed on a line.
     *  @param  useOffset                       if true, display the offset rather than the index.
     *  @return String                          - the dump.
     */
    public static String dump( byte[] buffer, int offset, int size, int width, boolean useOffset ){
        StringBuffer result = new StringBuffer( 1000 );
        for ( int index = 0, count = width;  index < size;  index++ ){
            if ( ++count >= width ){
                result.append( Log.endOfLine + String.format( "%04X", useOffset ? offset : index ) );
                count = 0;
            }
            result.append( " " + String.format( "%02X", buffer[ offset++ ] ) );
        }
        return result.toString();
    } // SerialIO.dump

} // class SerialIO
